/*
 * Copyright 2020 <copyright holder> <email>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef TRANSSECSWRAPPER_TRANSSECSTOOLWRAPPER_H
#define TRANSSECSWRAPPER_TRANSSECSTOOLWRAPPER_H

#include <jni.h>
#include "valueobject.h"
#include <map>

namespace transsecswrapper
{
    
    /**
     * @todo write docs
     */
    class TransSecsToolWrapper
    {
    public:
        /**
         * Default constructor
         * 
         * @param class_path 
         */
        TransSecsToolWrapper(const char* class_path);
        
        /**
         * Destructor. Note that this destructor will also destroy the JVM, which will invalidate any existing value objects and cause attempting to access
         * them to throw an error.
         */
        ~TransSecsToolWrapper();
        
        void testTopicBroker();
        
        /** Return the latest value associated with a topic.  This will return null if no value has been set. 
         * 
         * @param topic the topic
         */
        ValueObject Get(const char* topic);
        
        /** Publish an int.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, int value);
        
        /** Publish a long.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, long value);
        
        /** Publish a short.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, short value);
        
        /** Publish a float.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, float value);
        
        /** Publish a double.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, double value);
        
        /** Publish a string.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, std::string value);
        
        /** Publish a string.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, const char* value);
        
        /** Publish a char.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, char value);
        
        /** Publish a bool.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void Publish(const char* topic, bool value);
        
        /** Publish an object as json.
         * 
         * @param topic the topic to which to publish
         * @param value the value to publish
         */
        void PublishJson(const char* topic, const char* value);
        
        /** Add a callback for a topic. This is not guarenteed to be threadsafe; attempting to subscribe 
         * simultaniously in multiple threads could cause undefined behavior.
         * 
         * @param topic the topic
         * @param callback the call back
         */
        void Subscribe(const char* topic, void (*callback)(std::string, ValueObject));
        
        /** Remove a callback for a topic.
         * 
         * @param topic the topic
         * @param callback the call back
         */
        void Unsubscribe(const char* topic, void (*callback)(std::string, ValueObject));
        
        /** Set or clear an alarm.
         * 
         * @param alarmName
         */
        void Alarm(const char* alarm_name, bool set);
        
        /** Trigger the event with the DVIDs. 
         * 
         * @param eventName
         * @param ceidTrigger
         */
        void Trigger(const char* event_name, const char* ceid_trigger);
        
        /** Creates the tool
         */
        void StartMain(const char* tool_name);
    private:
        JavaVM *jvm;       /* denotes a Java VM */
        jclass broker_clazz;
        jobject topic_broker_instance;
        jmethodID get_topic_mid;
        std::map<std::pair<const char*, void (*)(std::string, ValueObject)>, jobject> listener_map;
    };
    
}

#endif // TRANSSECSWRAPPER_TRANSSECSTOOLWRAPPER_H
