
#include "../include/transsecstoolwrapper.h"
#include <stdexcept>
#include <stdio.h>
#include "../include/valueobject.h"
#include "com_ergotech_vib_servers_nativewrapper_NativeTopicListener.h"
#include <algorithm>

using namespace transsecswrapper;

TransSecsToolWrapper::TransSecsToolWrapper(const char* class_path)
{
    JNIEnv *env;       /* pointer to native method interface */
    JavaVMInitArgs vmArgs; /* JDK/JRE 8 VM initialization arguments */
    JavaVMOption* options = new JavaVMOption[1];
    std::string classpath_option_string = (std::string("-Djava.class.path=") + std::string(class_path));
    options[0].optionString = (char*)classpath_option_string.c_str();
    vmArgs.version = JNI_VERSION_1_8;
    vmArgs.nOptions = 1;
    vmArgs.options = options;
    vmArgs.ignoreUnrecognized = false;
    /* load and initialize a Java VM, return a JNI interface
     * pointer in env */
    int retval;
    if ((retval = JNI_CreateJavaVM(&jvm, (void **) &env, &vmArgs)) != JNI_OK) {
        throw std::runtime_error("Error when creating JVM");
    }
    delete[] options;
    broker_clazz = (jclass)env->NewGlobalRef(env->FindClass("com/ergotech/vib/servers/nativewrapper/TopicBroker"));
    jmethodID get_instance_mid = env->GetStaticMethodID(broker_clazz, "getInstance", "()Lcom/ergotech/vib/servers/nativewrapper/TopicBroker;");
    topic_broker_instance = env->NewGlobalRef(env->CallStaticObjectMethod(broker_clazz, get_instance_mid));
    get_topic_mid = env->GetMethodID(broker_clazz, "get", "(Ljava/lang/String;)Lcom/ergotech/vib/valueobjects/ValueObjectInterface;");
    
    // Register the listener class native method
    jclass native_listener_class = env->FindClass("com/ergotech/vib/servers/nativewrapper/NativeTopicListener");
    JNINativeMethod native_listener_native_method;
    native_listener_native_method.fnPtr = (void*) &Java_com_ergotech_vib_servers_nativewrapper_NativeTopicListener_topicValueChanged;
    native_listener_native_method.name = (char*)"topicValueChanged";
    native_listener_native_method.signature = (char*)"(Ljava/lang/String;Lcom/ergotech/vib/valueobjects/ValueObjectInterface;)V";
    env->RegisterNatives(native_listener_class, &native_listener_native_method, 1);
    jvm->DetachCurrentThread();
}

TransSecsToolWrapper::~TransSecsToolWrapper()
{
    // Don't need to do any cleanup because the whole jvm is getting destroyed
    jvm->DestroyJavaVM();
}


void transsecswrapper::TransSecsToolWrapper::testTopicBroker()
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID test_data_mid = env->GetStaticMethodID(broker_clazz, "addTestData", "()V");
    env->CallStaticVoidMethod(broker_clazz, test_data_mid);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

transsecswrapper::ValueObject transsecswrapper::TransSecsToolWrapper::Get(const char* tag)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    jobject get_object = env->CallObjectMethod(topic_broker_instance, get_topic_mid, key);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
    ValueObject v_object = ValueObject(jvm, get_object);
    return v_object;
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, int value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;J)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, bool value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;Z)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, char value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;J)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, double value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;D)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, float value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;D)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, long value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;J)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, short value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;J)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, const char* value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publish", "(Ljava/lang/String;Ljava/lang/String;)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    jstring java_value = env->NewStringUTF(value);
    env->CallVoidMethod(topic_broker_instance, publish_topic_mid, key, java_value);
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Publish(const char* tag, std::string value)
{
    Publish(tag, value.c_str());
}

void transsecswrapper::TransSecsToolWrapper::PublishJson(const char* tag, const char* value)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID publish_topic_mid = env->GetMethodID(broker_clazz, "publishObject", "(Ljava/lang/String;Ljava/lang/String;)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    jstring java_value = env->NewStringUTF(value);
    env->CallVoidMethod(topic_broker_instance, publish_topic_mid, key, java_value);
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}


void transsecswrapper::TransSecsToolWrapper::Subscribe(const char* tag, void (*callback)(std::string, transsecswrapper::ValueObject) )
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jclass native_listener_class = env->FindClass("com/ergotech/vib/servers/nativewrapper/NativeTopicListener");
    jmethodID native_listener_constructor = env->GetMethodID(native_listener_class, "<init>", "()V");
    jobject native_listener = env->NewGlobalRef(env->NewObject(native_listener_class, native_listener_constructor));
    jmethodID native_listener_get_uuid_mid = env->GetMethodID(native_listener_class, "getListenerUUID", "()[B");
    jbyteArray uuid_byte_array = (jbyteArray)env->CallObjectMethod(native_listener, native_listener_get_uuid_mid);
    jboolean is_copy;
    signed char* uuid_bytes = env->GetByteArrayElements(uuid_byte_array, &is_copy);
    long long uuid = *reinterpret_cast<unsigned long long*>(uuid_bytes);
    if (is_copy == JNI_TRUE) {
        env->ReleaseByteArrayElements(uuid_byte_array,uuid_bytes,JNI_ABORT); // JNI_ABORT -> release without copying back changes
    }
    
    listener_map[{tag, callback}] = native_listener;
    AddTopicListenerInternal(uuid, callback);
    
    jmethodID subscribe_topic_mid = env->GetMethodID(broker_clazz, "subscribe", "(Ljava/lang/String;Lcom/ergotech/vib/servers/nativewrapper/TopicListener;)V");
    
    jstring key = env->NewStringUTF(tag);

    env->CallVoidMethod(topic_broker_instance, subscribe_topic_mid, key, native_listener);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Unsubscribe(const char* tag, void (*callback)(std::string, transsecswrapper::ValueObject) )
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    
    jmethodID unsubscribe_topic_mid = env->GetMethodID(broker_clazz, "removeListener", "(Ljava/lang/String;Lcom/ergotech/vib/servers/nativewrapper/TopicListener;)V");
    std::string tag_string = std::string(tag);
    const char* chars = tag_string.c_str();
    jstring key = env->NewStringUTF(chars);
    env->CallVoidMethod(topic_broker_instance, unsubscribe_topic_mid, key, listener_map[{tag, callback}]);
    env->DeleteGlobalRef(listener_map[{tag, callback}]);
    listener_map.erase({tag, callback});
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::StartMain(const char* tool_name)
{
    JNIEnv* env;
    int get_env_stat = jvm->GetEnv((void **)&env, JNI_VERSION_1_8);
    bool is_detached = get_env_stat == JNI_EDETACHED;
    if (is_detached) {
        if (jvm->AttachCurrentThread((void **) &env, NULL) != 0) {
            throw std::runtime_error("Failed to AttachCurrentThread");
        }
    } else if (get_env_stat == JNI_EVERSION) {
        throw std::runtime_error("GetEnv: version not supported");
    } else if (get_env_stat != JNI_OK) {
        throw std::runtime_error("JNI error");
    }
    std::string class_name = (std::string("deploy/")+std::string(tool_name)+std::string("/EquipmentController"));
    jclass equipment_controller_class = env->FindClass(class_name.c_str());
    jmethodID start_main_mid = env->GetStaticMethodID(equipment_controller_class, "startMain", "()V");
    env->CallStaticVoidMethod(equipment_controller_class, start_main_mid);
    
    if (is_detached) {
        jvm->DetachCurrentThread();
    }
}

void transsecswrapper::TransSecsToolWrapper::Alarm ( const char* alarm_name, bool set )
{
    std::string alarm_name_string = std::string(alarm_name);
    // convert the string to lowercase
    std::transform(alarm_name_string.begin(), alarm_name_string.end(), alarm_name_string.begin(),
        [](unsigned char c){ return std::tolower(c); });
    Publish((std::string("gemtool/ceids/") + alarm_name_string + std::string("/trigger")).c_str(), set);
}

void transsecswrapper::TransSecsToolWrapper::Trigger ( const char* event_name, const char* ceid_trigger )
{
    std::string event_name_string = std::string(event_name);
    // convert the string to lowercase
    std::transform(event_name_string.begin(), event_name_string.end(), event_name_string.begin(),
        [](unsigned char c){ return std::tolower(c); });
    Publish((std::string("gemtool/ceids/") + event_name_string + std::string("/trigger")).c_str(), ceid_trigger);
}


